unit MSL;

interface

uses Windows, PEUtils, SysUtils, MSLTypes;

function LoadLib(LibFileName: PChar):HMODULE; stdcall;
function FreeLib(HMODULE: hModule):LONGBOOL; stdcall;
function GetProcAddr(HMODULE: hModule; ProcName: PChar):
 FARPROC; stdcall;
function MappedDllToMsl(DLLOffset: Pointer; PSize: PDWORD):
 Pointer; stdcall;
function DllToMsl(DllFileName, MslFileName: PChar): LONGBOOL;
 stdcall;
function LoadMappedMsl(Base: DWORD): DWORD; stdcall;

function GetMslSection(H: PMslHeader; i: integer):PMslSectionHeader;

implementation

function GetMslSection(H: PMslHeader; i: integer):PMslSectionHeader;
begin
  Result:=Ptr(DWORD(H)+SizeOf(TMslHeader)+
   (DWORD(i)-1)*SizeOf(TMslSectionHeader));
end;

procedure ProcessImports(ImageBase, OrgIT: DWORD);
var OrgFirstThunk, Name1, Buffer, FirstThunk,
    OrgThunkData: DWORD;
begin
  asm
    pushad
    mov edx, OrgIT
    add edx, ImageBase
    mov eax,[edx]
    add eax,ImageBase
    mov OrgFirstThunk,eax
    add edx,12
    mov eax,[edx]
    add eax, ImageBase
    mov Name1,eax
    add edx,4
    mov eax,[edx]
    add eax, ImageBase
    mov FirstThunk,eax
    add edx,4
    mov Buffer,edx

    //*****Address loader*****//
   @MainITLoop:
    mov ecx,ImageBase
    cmp FirstThunk,ecx
    jz @MainITEnd

    mov ebx,Name1
    push ebx
    //mov eax,dword [ebp+@_LoadLibrary]
    //call eax
    call LoadLibraryA
    mov ebx,eax

    mov ecx, ImageBase
    cmp OrgFirstThunk,ecx
    jnz @Nothing
    mov eax,FirstThunk
    mov OrgFirstThunk,eax

   @Nothing:
    mov ecx,OrgFirstThunk
    mov eax,[ecx]
    mov OrgThunkData,eax

   @SecondITLoop:
    cmp OrgThunkData,0
    jz @OrgThunkJMP
    mov eax,OrgThunkData
    and eax,$80000000  // IMAGE_ORDINAL_FLAG32
    cmp eax,0
    jnz @Ordinal

    mov ecx,OrgThunkData
    mov edx,ImageBase
    add edx,2
    add ecx,edx
    mov edi,ecx
    push edi
    push ebx
    //mov eax,dword [ebp+@_GetProcAddress]
    //call eax
    call GetProcAddress
    jmp @Jump

   @Ordinal:
    mov eax,OrgThunkData
    sub eax,80000000h
    push eax
    push ebx
    //mov eax,dword [ebp+@_GetProcAddress]
    //call eax
    call GetProcAddress

   @Jump:
    mov esi,FirstThunk
    mov dword ptr [esi],eax
    mov ecx,OrgFirstThunk
    mov edx,FirstThunk
    add ecx,4
    add edx,4
    mov OrgFirstThunk,ecx
    mov FirstThunk,edx
    mov eax,[ecx]
    mov OrgThunkData,eax
    jmp @SecondITLoop

   @OrgThunkJMP:
    mov edx,Buffer
    mov eax,[edx]
    add eax,ImageBase
    mov OrgFirstThunk,eax
    add edx,12
    mov eax,[edx]
    add eax,ImageBase
    mov Name1,eax
    add edx,4
    mov eax,[edx]
    add eax,ImageBase
    mov FirstThunk,eax
    add edx,4
    mov Buffer,edx
    jmp @MainITLoop

   @MainITEnd:
    popad
  end;
end;

procedure HandleReloc(OldImageBase, NewImageBase, Reloc: DWORD);
begin
  asm
    pushad
    mov edx, NewImageBase
    mov eax, OldImageBase
    sub edx, eax		 // Calculate Delta
    jz @EndReloc 		 // if zero =No relocation problem

    mov eax, edx
    shr eax, 16			 // AX= high 16 bit

    xor ebx, ebx		 // Prepare EBX

    mov esi, Reloc
    add esi, NewImageBase        // ESI = reloctable

   @NextPage:			 // Do a page
    cmp	dword ptr [esi],0	 // are we done?
    jz 	@EndReloc

    mov ecx, [ESI+4]		 // number of bytes in Structure
    sub ecx, 8			 // minus RVA & Size Entries
    shr ecx, 1			 // 1 word per entry
				 // => ecx= number of entries

    mov edi, [esi]		 // Current Page RVA
    add edi, NewImageBase        // RVA=>VA

    add esi, 8			 // ESI=FirstItem Entry
   @NextItem:
    mov bx, word ptr [ESI]
    shr	ebx, 12			 // EBX= type

    cmp	ebx, 1
    jz 	@type1

    cmp	ebx,2
    jz 	@type2

    cmp	ebx, 3
    jz @type3

    jmp @NoType

   @type1:
    mov	bx, word ptr [ESI]
    and ebx, 0fffh		  // 8fffh right mask?
    add word ptr [edi+ebx], ax
    jmp @NoType
   @type2:
    mov	bx, word ptr [ESI]
    and ebx, 0fffh

    add word ptr [edi+ebx], dx
    jmp @NoType
   @type3:
    mov	bx, word ptr [ESI]
    and ebx, 0fffh
    add dword ptr [edi+ebx],edx

    jmp @NoType

   @NoType:
    add esi, 2		// next offset
    loop @NextItem

    jmp @NextPage

   @EndReloc:
    popad
  end;
end;

function SectionSize(const Section: PImageSectionHeader):DWORD;
// gets the Section's size
begin
  if Section.SizeOfRawData<Section.Misc.VirtualSize then
     SectionSize:=Section.SizeOfRawData
  else
      SectionSize:=Section.Misc.VirtualSize;
end;

function GetProcAddr(HMODULE: hModule; ProcName: PChar):FARPROC;
var ExpDir: PIMAGE_EXPORT_DIRECTORY;
    H: PImageNtHeaders;
    cIndex: DWORD;
    CurrProcName: PChar;
begin
  Result:=nil;
  H:=GetImageNTHeaders(HMODULE);
  if H=nil then exit;
  ExpDir:=Ptr(HMODULE+H.OptionalHeader.DataDirectory[0].VirtualAddress);

  // if it's Ordinal import then High Word is zero
  if (PDWORD(ProcName)^ > $FFFF) then
  begin
    for cIndex:=0 to ExpDir.NumberOfNames-1 do
    begin
      CurrProcName:=Ptr(PDWORD(Ptr(ExpDir.AddressOfNames+
       HMODULE+cIndex*4))^+HMODULE);
      if String(CurrProcName)=String(ProcName) then
      begin
        Result:=Ptr(PDWORD(Ptr(ExpDir.AddressOfFunctions+
         HMODULE+cIndex*4))^+HMODULE);
        //Result:=Ptr(ExpDir.AddressOfFunctions+HMODULE);
        //Result:=Ptr(PDWORD(Result)^+cIndex*4+HMODULE);
        exit;
      end;
    end;
  end
  else
      // Ordinal import - not implemented...
end;

function FreeLib(HMODULE: hModule):LONGBOOL;
var H: PImageNtHeaders;
begin
  Result:=False;
  H:=GetImageNTHeaders(HMODULE);
  if H=nil then Exit;

  if H.OptionalHeader.AddressOfEntryPoint<>0 then
  asm // call DllMain with DLL_PROCESS_DETACH
    push eax
    push ecx
    push edx
    mov eax, H
    mov eax, [eax].IMAGE_NT_HEADERS.OptionalHeader.AddressOfEntryPoint
    add eax, HMODULE
    push 0
    push DLL_PROCESS_DETACH
    push HMODULE
    call eax
    pop edx
    pop ecx
    pop eax
  end;

  VirtualFree(Ptr(HMODULE),H.OptionalHeader.SizeOfImage,MEM_DECOMMIT);
  Result:=True;
end;

function LoadMappedMsl(Base: DWORD): DWORD;
var Header: PMslHeader;
    Section: PMslSectionHeader;
    i: integer;
    ImageBase: Pointer;
    PEHdr: PImageNtHeaders;
begin
  Result:=0;
  Header:=Ptr(Base);
  if Header.Signature<>MSL_SIGNATURE_10 then
     Exit;

  // try to get memory on preferred ImageBase
  ImageBase:=VirtualAlloc(Ptr(Header.ImageBase),Header.SizeOfImage,
   MEM_COMMIT, PAGE_EXECUTE_READWRITE);

  // if Header.ImageBase was too low, VirtualAlloc didn't succeed
  // -> try again and let Windows decide where to allocate
  if ImageBase=nil then
     ImageBase:=VirtualAlloc(nil,Header.SizeOfImage,
      MEM_COMMIT, PAGE_EXECUTE_READWRITE);

  if ImageBase=nil then exit; // another error occured...

  // place every section on it's VirtualAddress (map them)
  for i:=1 to Header.NumberOfSections do
  begin
    Section:=GetMslSection(Header, i);
    CopyMemory(Ptr(DWORD(ImageBase)+Section.VirtualAddress),
      Ptr(Base+Section.RawAddress),Section.RawSize);
  end;

  // Build up MZ - PE header and fill in some values...
  PWORD(ImageBase)^:=$5A4D; // 'MZ'
  PWORD(Ptr(DWORD(ImageBase)+$3C))^:=$C; // PE offset
  PWORD(Ptr(DWORD(ImageBase)+$C))^:=$4550; // 'PE'
  PEHdr:=Ptr(DWORD(ImageBase)+PWORD(Ptr(DWORD(ImageBase)+$3C))^);
  PEHdr.OptionalHeader.AddressOfEntryPoint:=Header.EntryPoint;
  PEHdr.OptionalHeader.SizeOfImage:=Header.SizeOfImage;
  PEHdr.OptionalHeader.DataDirectory[0].VirtualAddress:=Header.ExportsVA;
  PEHdr.OptionalHeader.DataDirectory[2].VirtualAddress:=Header.ResourcesVA;

  // Fixups needed to make FindResource & co. working
  PEHdr.OptionalHeader.Magic:=$10B;
  PEHdr.OptionalHeader.NumberOfRvaAndSizes:=IMAGE_NUMBEROF_DIRECTORY_ENTRIES;
  PEHdr.OptionalHeader.DataDirectory[2].Size:=Header.SizeOfImage-
   Header.ResourcesVA; // just approx. size - no one needs a correct one

  // enables 3D look of controls (on dialogs, etc.)
  PEHdr.OptionalHeader.MajorSubsystemVersion:=4;
  PEHdr.OptionalHeader.MinorSubsystemVersion:=0;

  // Process Imported functions (fill the import table
  // with addresses of imported functions)
  ProcessImports(DWORD(ImageBase), Header.ImportsVA);

  // and handle relocations...
  HandleReloc(Header.ImageBase,DWORD(ImageBase),Header.RelocationsVA);

  asm
    //int 3
    push eax
    push ecx
    push edx // these registers aren't preseved, others are
             // according to StdCall preserved

    mov eax, PEHdr
    mov eax, [eax].IMAGE_NT_HEADERS.OptionalHeader.AddressOfEntryPoint
    add eax, ImageBase
    // BOOL APIENTRY DllMain( HANDLE hModule,
    //	DWORD dwReason,
    //	LPVOID lpReserved )
    push 0
    push DLL_PROCESS_ATTACH
    push ImageBase
    call eax
    cmp eax, 1 // should return 1 if successfull
    jne @LeaveUs
    mov eax, ImageBase
    mov Result, eax
   @LeaveUs:
    pop edx
    pop ecx
    pop eax
  end;
end;

function LoadLib(LibFileName: PChar):HMODULE;
var M: Pointer;
    Size: DWORD;
begin
  M:=LoadFile(String(LibFileName),Size);
  Result:=LoadMappedMsl(DWORD(M));
  FreeMem(M); // free the memory occupied by file
end;

function MappedDllToMsl(DLLOffset: Pointer; PSize: PDWORD): Pointer;
var H: PImageNtHeaders;
    S: PImageSectionHeader;
    Header: PMslHeader;
    Section: PMslSectionHeader;
    i: integer;
    CurrHeaderOffset: DWORD;
    CurrFileOffset: DWORD;
begin
  Result:=nil;
  H:=GetImageNTHeaders(DWORD(DllOffset));
  if H=nil then exit;

  PSize^:=SizeOf(TMslHeader);
  for i:=1 to H.FileHeader.NumberOfSections do
      PSize^:=PSize^+SectionSize(GetSection(H, i))+
       SizeOf(TMslSectionHeader);
  GetMem(Result, PSize^);

  // create MSL header
  Header:=Result;
  Header.Signature:=MSL_SIGNATURE_10;
  Header.ImageBase:=H.OptionalHeader.ImageBase;
  Header.EntryPoint:=H.OptionalHeader.AddressOfEntryPoint;
  Header.SizeOfImage:=H.OptionalHeader.SizeOfImage;
  Header.NumberOfSections:=H.FileHeader.NumberOfSections;
  Header.ImportsVA:=H.OptionalHeader.DataDirectory[1].VirtualAddress;
  Header.ExportsVA:=H.OptionalHeader.DataDirectory[0].VirtualAddress;
  Header.RelocationsVA:=H.OptionalHeader.DataDirectory[5].VirtualAddress;
  Header.ResourcesVA:=H.OptionalHeader.DataDirectory[2].VirtualAddress;

  // copy all sections to their positions
  CurrHeaderOffset:=SizeOf(TMslHeader);
  CurrFileOffset:=SizeOf(TMslHeader)+
   SizeOf(TMslSectionHeader)*Header.NumberOfSections;
  for i:=1 to Header.NumberOfSections do
  begin
    Section:=Ptr(DWORD(Result)+CurrHeaderOffset);
    S:=GetSection(H, i);
    Section.RawAddress:=CurrFileOffset;
    Section.RawSize:=SectionSize(S);
    Section.VirtualAddress:=S.VirtualAddress;
    CopyMemory(Ptr(DWORD(Result)+CurrFileOffset),
     Ptr(DWORD(DllOffset)+S.PointerToRawData),
     SectionSize(S));
    inc(CurrFileOffset, SectionSize(S));
    inc(CurrHeaderOffset, SizeOf(TMslSectionHeader));
  end;
end;

function DllToMsl(DllFileName, MslFileName: PChar): LONGBOOL;
var Dll, Msl: Pointer;
    Size: DWORD;
begin
  Dll:=LoadFile(String(DllFileName),Size);
  Msl:=MappedDllToMsl(Dll, @Size);
  if Msl=nil then
     Result:=False
  else
  begin
    SaveFile(Msl, Size, String(MslFileName));
    FreeMem(Msl);
    Result:=True;
  end;
  FreeMem(Dll);
end;


end.
