unit PEUtils;

interface

// migeel's PE (Portable Executable) manipulation routines
// Copyright (c) 2003-2004, Michal Strehovsky
// www.migeel.sk
// Comments? -> mail@migeel.sk

uses Windows, ImageHlp, SysUtils;

function LoadFile(FileName:String; var Size: DWORD):Pointer;
procedure SaveFile(M: Pointer; Size: DWORD; FileName: String);
function GetImageNTHeaders(Address: DWORD):PImageNtHeaders;
function GetSection(NTHeader: PImageNtHeaders; Section: word):PImageSectionHeader;
function PEAlign(Num, AlignTo: DWORD):DWORD;
function AddPESection(m: Pointer; virSize,Characteristics:DWORD):PImageSectionHeader;
function RVAToSection(Head: PImageNtHeaders; RVA: DWORD):PImageSectionHeader;
function RAWToSection(Head: PImageNtHeaders; RAW: DWORD):PImageSectionHeader;
function RVAToRAW(Head: PImageNtHeaders; RVA: DWORD):DWORD;
function RAWToRVA(Head: PImageNtHeaders; RAW: DWORD):DWORD;
procedure ProtectAgainstIDA(Head: PImageNtHeaders);
procedure ObfuscatePE(Head: PImageNtHeaders);
function FindCave(Head: PImageNtHeaders):PImageSectionHeader;
function StripRelocs(M: Pointer; var Size: DWORD):boolean;

type TDWORDArray = array [1..8191] of dword;
     PDWORDArray = ^TDWORDArray;

implementation

function LoadFile(FileName:String; var Size: DWORD):Pointer;
// This loads file into memory, returns pointer to it and the filesize
var hFile: THandle;
    Position, BRead: DWORD;
    Buffer: array[1..8192] of char;
begin
  hFile:=FileOpen(FileName,fmOpenRead or fmShareDenyNone);

  if hFile=INVALID_HANDLE_VALUE then
  begin
    Size:=0;
    Result:=nil;
    exit;
  end;

  // Allocate Memory for file
  Size:=GetFileSize(hFile,nil);
  GetMem(Result, Size);
  Position:=0;
  BRead:=FileRead(hFile,Buffer,SizeOf(Buffer));
  while BRead=SizeOf(Buffer) do
  begin
    CopyMemory(Ptr(DWORD(Result)+Position),@Buffer,BRead);
    inc(Position,BRead);
    BRead:=FileRead(hFile,Buffer,SizeOf(Buffer));
  end;
  if BRead>0 then CopyMemory(Ptr(DWORD(Result)+Position),@Buffer,BRead);
  FileClose(hFile);
end;

procedure SaveFile(M: Pointer; Size: DWORD; FileName: String);
// saves memory area to file
var hFile: THandle;
    BRead, Position: DWORD;
    Buffer: array [1..8192] of byte;
begin
  hFile:=FileCreate(FileName);
  Position:=0;
  while Position<Size do
  begin
    BRead:=Size-Position;
    if BRead>SizeOf(Buffer) then BRead:=SizeOf(Buffer);
    CopyMemory(@Buffer,Ptr(DWORD(M)+Position),BRead);
    Inc(Position,BRead);
    FileWrite(hFile,Buffer,BRead);
  end;
  FileClose(hFile);
end;

function GetImageNTHeaders(Address: DWORD):PImageNtHeaders;
var NewHeader: PWord;
begin
  NewHeader:=Ptr(Address+$3C);     // Position of PE header
  result:=Ptr(NewHeader^+Address); // Map it to IMAGE_NT_HEADERS
  if PDWORD(Result)^<>IMAGE_NT_SIGNATURE then
     result:=nil;
end;

function GetSection(NTHeader: PImageNtHeaders; Section: word):PImageSectionHeader;
var Adr: DWORD;
begin
  Adr:=Integer(NTHeader)+SizeOf(IMAGE_NT_HEADERS)+(Section-1)*SizeOf(IMAGE_SECTION_HEADER);
  Result:=Ptr(Adr);
end;

function PEAlign(Num, AlignTo: DWORD):DWORD;
begin
  while (Num mod AlignTo)<>0 do
        inc(Num);
  Result:=Num;
end;

function AddPESection(m: Pointer; VirSize,Characteristics:DWORD):PImageSectionHeader;
var NewHdrSize, HdrSize: DWORD;
    NewSection, pSection: PImageSectionHeader;
    Headers: PImageNtHeaders;
begin
  Headers:=GetImageNTHeaders(DWORD(m));

  NewHdrSize:=Headers.FileHeader.NumberOfSections*SizeOf(IMAGE_SECTION_HEADER);
  HdrSize:=Dword(Headers)-Dword(m)+NewHdrSize+SizeOf(IMAGE_NT_HEADERS);
  if HdrSize>Headers.OptionalHeader.SizeOfHeaders then
  begin
    result:=nil;
    exit;
  end;

  pSection:=GetSection(Headers,Headers.FileHeader.NumberOfSections);
  NewSection:=GetSection(Headers,Headers.FileHeader.NumberOfSections+1);
  NewSection.Name[0]:=Ord('.');
  NewSection.Name[1]:=Ord('m');
  NewSection.Name[2]:=Ord('i');
  NewSection.Name[3]:=Ord('g');
  NewSection.Name[4]:=Ord('e');
  NewSection.Name[5]:=Ord('e');
  NewSection.Name[6]:=Ord('l');
  NewSection.Name[7]:=0;        // cool section name, isn't it? ;-)

  NewSection.VirtualAddress:=PEAlign(pSection.VirtualAddress+
        pSection.Misc.VirtualSize,Headers.OptionalHeader.SectionAlignment);
  NewSection.Misc.VirtualSize:=VirSize;
  NewSection.SizeOfRawData:=PEAlign(VirSize,Headers.OptionalHeader.FileAlignment);
  NewSection.Characteristics:=Characteristics;
  NewSection.PointerToRawData:=pSection.PointerToRawData+pSection.SizeOfRawData;

  inc(Headers.FileHeader.NumberOfSections);
  inc(Headers.OptionalHeader.SizeOfImage,NewSection.SizeOfRawData);
  inc(Headers.OptionalHeader.SizeOfCode,NewSection.SizeOfRawData);
  inc(Headers.OptionalHeader.SizeOfInitializedData,NewSection.SizeOfRawData);

  Result:=NewSection;
end;

function RVAToSection(Head: PImageNtHeaders; RVA: DWORD):PImageSectionHeader;
// this is because ImageRVAToSection exported from
// imagehlp.dll sucks. Sucks a lot. And doesn't work...
var i: integer;
    Section: PImageSectionHeader;
begin
  Section:=nil;
  for i:=1 to Head.FileHeader.NumberOfSections do
  begin
    Section:=GetSection(Head, i);
    if Section.VirtualAddress>RVA then
    begin
      result:=GetSection(Head, i-1);
      exit;
    end;
  end;
  Result:=Section; // most probably the last section...
end;

function RAWToSection(Head: PImageNtHeaders; RAW: DWORD):PImageSectionHeader;
var i: integer;
    Section: PImageSectionHeader;
begin
  Section:=nil;
  for i:=1 to Head.FileHeader.NumberOfSections do
  begin
    Section:=GetSection(Head, i);
    if Section.PointerToRawData>RAW then
    begin
      result:=GetSection(Head, i-1);
      exit;
    end;
  end;
  Result:=Section; // most probably the last section...
end;

function RVAToRAW(Head: PImageNtHeaders; RVA: DWORD):DWORD;
var s: PImageSectionHeader;
begin
  s:=RVAToSection(Head, RVA);
  Result:=RVA-s.VirtualAddress+s.PointerToRawData;
end;

function RAWToRVA(Head: PImageNtHeaders; RAW: DWORD):DWORD;
var s: PImageSectionHeader;
begin
  s:=RAWToSection(Head, RAW);
  Result:=RAW-s.PointerToRawData+s.VirtualAddress;
end;

procedure ProtectAgainstIDA(Head: PImageNtHeaders);
// a nice trick against the Interactive Disassembler ;-)
begin
  // this will point exports to OEP :D
  // OEP is stored in TimeDateStamp
  if Head.OptionalHeader.DataDirectory[0].VirtualAddress=0 then
  begin
    Randomize;
    Head.OptionalHeader.DataDirectory[0].VirtualAddress:=
     Head.FileHeader.TimeDateStamp;
    Head.OptionalHeader.DataDirectory[0].Size:=0{Round(Random*40+1)};
  end;
end;

procedure ObfuscatePE(Head: PImageNtHeaders);
var i: integer;
    Section: PImageSectionHeader;
begin
  for i:=1 to Head.FileHeader.NumberOfSections-1 do
  begin
    Section:=GetSection(Head,i);

    // increase section's size (range from +1 to +$1FF)
    // only if it's not the resource section
    if String(@Section.Name)<>'.rsrc' then
       Section.PointerToRawData:=Section.PointerToRawData+$100{Round(Random*$1FE)+1};
  end;
end;

function FindCave(Head: PImageNtHeaders):PImageSectionHeader;
var i: integer;
    Biggest, S: PImageSectionHeader;
begin
  Biggest:=GetSection(Head, 1);
  for i:=1 to Head.FileHeader.NumberOfSections do
  begin
    S:=GetSection(Head, i);

    // prevent overflow
    if S.Misc.VirtualSize>S.SizeOfRawData then Continue;

    if (S.SizeOfRawData-S.Misc.VirtualSize)>
       (Biggest.SizeOfRawData-Biggest.Misc.VirtualSize) then
       Biggest:=S;
  end;
  Result:=Biggest;
end;

function StripRelocs(M: Pointer; var Size: DWORD):boolean;
const RelocSectionName = '.reloc';
var H: PImageNtHeaders;
    i: integer;
    RelocPhysOffset, RelocPhysSize: DWORD;
    S: PImageSectionHeader;
begin
  Result:=False;
  H:=GetImageNTHeaders(DWORD(M));

  // if ImageBase < $400000 the EXE WILL SURELY BE
  // RELOCATED by Windows PE loader
  if H.OptionalHeader.ImageBase<$400000 then exit;

  if (H.FileHeader.Characteristics and IMAGE_FILE_RELOCS_STRIPPED)<>0 then
     exit;
  H.FileHeader.Characteristics:=H.FileHeader.Characteristics or IMAGE_FILE_RELOCS_STRIPPED;

  if (H.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress=0) or
     (H.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size=0) then
     exit;
  H.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress:=0;
  H.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size:=0;

  // Find the '.reloc' section
  RelocPhysOffset:=0;
  for i:=1 to H.FileHeader.NumberOfSections do
  begin
    S:=GetSection(H, i);
    if StrPas(@S.Name[0])=RelocSectionName then
    begin
      RelocPhysOffset:=S.PointerToRawData;
      RelocPhysSize:=S.SizeOfRawData;
      S.PointerToRawData:=0;
      S.SizeOfRawData:=0;
      Break;
    end;
  end;
  if RelocPhysOffset=0 then Exit;

  // Move PointerToRawData of sections which are AFTER
  // the .reloc section
  for i:=1 to H.FileHeader.NumberOfSections do
  begin
    S:=GetSection(H, i);
    if S.PointerToRawData>=RelocPhysOffset then
       Dec(S.PointerToRawData,RelocPhysSize);
    if S.PointerToLinenumbers>=RelocPhysOffset then
       Dec(S.PointerToLinenumbers,RelocPhysSize);
  end;

  // and move the sections after the .reloc section
  CopyMemory(Ptr(RelocPhysOffset+DWORD(M)),
   Ptr(RelocPhysOffset+DWORD(M)+RelocPhysSize),
   Size-RelocPhysOffset-RelocPhysSize);

  Size:=Size-RelocPhysSize;
  Result:=True;
end;


begin

end.
